@@ -11,12 +11,16 @@ import itertools
 import subprocess
 import igraph as ig
 import leidenalg as la
+from contextlib import closing
+from multiprocessing import Pool
+from itertools import repeat
 from numba import njit, float64
 from scipy import sparse as sp
 from scipy.linalg import norm
 from sklearn.preprocessing import normalize
 from phenograph import bruteforce_nn # temproarly using phenograph on platforms other than Linux
 
+
 if __name__ == '__main__':
     def main():
         pass
@@ -57,20 +61,25 @@ def minus_1_normalize_arcsinh(X, noise_threshold = 1, base = 5):
     temp[np.where(np.all(np.isclose(temp, 0), axis=1))[0]] = 1.0
     return(normalize_arcsinh(temp, base=base))
 
-def preprocessing(X, method = None):
+def preprocessing(X, platform = None):
     temp = type(X)
-    if method == 'mass':
+    if platform == 'mass':
         if temp is pd.core.frame.DataFrame:
             X = normalize_arcsinh(X.select_dtypes([np.number]).to_numpy(), base = 5)
         else:
             X = normalize_arcsinh(X, base = 5)
-    elif method == 'flow':
+    elif platform == 'flow':
         if temp is pd.core.frame.DataFrame:
             X = normalize_arcsinh(X.select_dtypes([np.number]).to_numpy(), base = 150)
         else:
             X = normalize_arcsinh(X, base = 150)
+    elif platform == 'spectral flow':
+        if temp is pd.core.frame.DataFrame:
+            X = normalize_arcsinh(X.select_dtypes([np.number]).to_numpy(), base = 6000)
+        else:
+            X = normalize_arcsinh(X, base = 6000)
             
-    elif method == 'scrnaseq':
+    elif platform == 'scrnaseq':
         import scanpy as sc
         import anndata
 
@@ -162,7 +171,8 @@ def build_graph(rows, columns, weights, shape):
     """
     warnings.filterwarnings('ignore')
     csr = sp.csr_matrix((np.array(weights), (np.array(rows), np.array(columns))), shape=shape)
-    csr = (csr + csr.transpose()).multiply(0.5)
+    #csr = (csr + csr.transpose()).multiply(0.5)
+    csr = csr+csr.transpose() - csr.multiply(csr.transpose())
     csr.setdiag(0)
     csr = sp.triu(csr)
     csr.eliminate_zeros()
@@ -174,7 +184,162 @@ def build_graph(rows, columns, weights, shape):
     return G
 
 
-def build_tani_graph(clu_file, data):
+
+SMOOTH_K_TOLERANCE = 1e-5
+MIN_K_DIST_SCALE = 1e-3
+NPY_INFINITY = np.inf
+
+def smooth_knn_dist(weights, shape, k, n_iter=64, local_connectivity=1.0, bandwidth=1.0):
+    """Compute a continuous version of the distance to the kth nearest
+    neighbor. That is, this is similar to knn-distance but allows continuous
+    k values rather than requiring an integral k. In essence we are simply
+    computing the distance such that the cardinality of fuzzy set we generate
+    is k.
+
+    Parameters
+    ----------
+    distances: array of shape (n_samples, n_neighbors)
+        Distances to nearest neighbors for each samples. Each row should be a
+        sorted list of distances to a given samples nearest neighbors.
+
+    k: float
+        The number of nearest neighbors to approximate for.
+
+    n_iter: int (optional, default 64)
+        We need to binary search for the correct distance value. This is the
+        max number of iterations to use in such a search.
+
+    local_connectivity: int (optional, default 1)
+        The local connectivity required -- i.e. the number of nearest
+        neighbors that should be assumed to be connected at a local level.
+        The higher this value the more connected the manifold becomes
+        locally. In practice this should be not more than the local intrinsic
+        dimension of the manifold.
+
+    bandwidth: float (optional, default 1)
+        The target bandwidth of the kernel, larger values will produce
+        larger return values.
+
+    Returns
+    -------
+    knn_dist: array of shape (n_samples,)
+        The distance to kth nearest neighbor, as suitably approximated.
+
+    nn_dist: array of shape (n_samples,)
+        The distance to the 1st nearest neighbor for each point.
+    """
+    target = np.log2(k) * bandwidth
+    distances = np.array(weights).reshape(shape)
+    rho = np.zeros(shape[0], dtype=np.float32)
+    result = np.zeros(shape[0], dtype=np.float32)
+
+    mean_distances = np.mean(distances)
+
+    for i in range(shape[0]):
+        lo = 0.0
+        hi = NPY_INFINITY
+        mid = 1.0
+
+        # TODO: This is very inefficient, but will do for now. FIXME
+        ith_distances = distances[i]
+        non_zero_dists = ith_distances[ith_distances > 0.0]
+        if non_zero_dists.shape[0] >= local_connectivity:
+            index = int(np.floor(local_connectivity))
+            interpolation = local_connectivity - index
+            if index > 0:
+                rho[i] = non_zero_dists[index - 1]
+                if interpolation > SMOOTH_K_TOLERANCE:
+                    rho[i] += interpolation * (
+                        non_zero_dists[index] - non_zero_dists[index - 1]
+                    )
+            else:
+                rho[i] = interpolation * non_zero_dists[0]
+        elif non_zero_dists.shape[0] > 0:
+            rho[i] = np.max(non_zero_dists)
+
+        for n in range(n_iter):
+
+            psum = 0.0
+            for j in range(1, distances.shape[1]):
+                d = distances[i, j] - rho[i]
+                if d > 0:
+                    psum += np.exp(-(d / mid))
+                else:
+                    psum += 1.0
+
+            if np.fabs(psum - target) < SMOOTH_K_TOLERANCE:
+                break
+
+            if psum > target:
+                hi = mid
+                mid = (lo + hi) / 2.0
+            else:
+                lo = mid
+                if hi == NPY_INFINITY:
+                    mid *= 2
+                else:
+                    mid = (lo + hi) / 2.0
+
+        result[i] = mid
+
+        # TODO: This is very inefficient, but will do for now. FIXME
+        if rho[i] > 0.0:
+            mean_ith_distances = np.mean(ith_distances)
+            if result[i] < MIN_K_DIST_SCALE * mean_ith_distances:
+                result[i] = MIN_K_DIST_SCALE * mean_ith_distances
+        else:
+            if result[i] < MIN_K_DIST_SCALE * mean_distances:
+                result[i] = MIN_K_DIST_SCALE * mean_distances
+
+    return result, rho
+
+
+def build_graph_umap(rows, columns, weights, shape, nbr_num, set_op_mix_ratio=1.0,):
+    """
+    Build a igraph.Graph from the pruned graph
+
+    Parameters
+    ----------
+    rows : list
+        list of row coordinates of the cells
+    columns : list
+        list of column coordinates of the cells
+    weights : list
+        list of Tanimoto coefficients as weights
+    shape : tuple
+        shape of the adjacency matrix, number of cells by number of cells
+
+    Returns
+    -------
+    igraph.Graph
+        converted graph to be the input of the Leiden algorithm
+    """
+    from umap import umap_ as u
+    
+    warnings.filterwarnings('ignore')
+    distances = np.array(weights).reshape(shape[0], nbr_num)
+    indices = np.array(columns).reshape(shape[0], nbr_num)
+    sigmas, rhos = u.smooth_knn_dist(distances, nbr_num)
+    r, c, v, d = u.compute_membership_strengths(indices, distances, sigmas, rhos, return_dists = False)
+    
+    result = sp.coo_matrix((v, (r, c)), shape=shape)
+    result.eliminate_zeros()
+
+    transpose = result.transpose()
+    prod_matrix = result.multiply(transpose)
+    result = (
+            set_op_mix_ratio * (result + transpose - prod_matrix)
+            + (1.0 - set_op_mix_ratio) * prod_matrix
+        )
+    result.eliminate_zeros()
+    warnings.resetwarnings()
+    sources, targets = result.nonzero()
+    edgelist = list(zip(sources.tolist(), targets.tolist()))
+    G = ig.Graph(edges = edgelist, edge_attrs = {'weight': result.data.tolist()}) 
+    #warnings.filterwarnings('default')
+    return G
+
+def build_tani_graph(clu_file, data, local = False, method = 'tani'):
     """
     readin the clu result file generated by l2knng and calculate the Tanimoto coefficients, then convert the result to an igraph.Graph as the input for the Leiden algorithm
 
@@ -213,15 +378,25 @@ def build_tani_graph(clu_file, data):
     cos_vals = list(map(float, list(itertools.chain.from_iterable(csr))[1::2]))
     columns = list(i - 1 for i in columns) ## bug in L2knng
 
-    if sp.issparse(data):
-        cell_norm = sp.linalg.norm(data, axis = 1)
+    if method =='tani':
+        if sp.issparse(data):
+            cell_norm = sp.linalg.norm(data, axis = 1)
+        else:
+            cell_norm = norm(data, axis = 1)
+        weights = [tani_from_cos(cell_norm[i],cell_norm[j],v) for i,j,v in zip(rows, columns, cos_vals)]
+    elif method =='jaccard':
+        nbr_index = np.array(columns).reshape(shape)
+        weights = np.ravel(map_jaccard(nbr_index)).tolist()
+    elif method =='cosine':
+        weights = cos_vals
     else:
-        cell_norm = norm(data, axis = 1)
-    weights = [tani_from_cos(cell_norm[i],cell_norm[j],v) for i,j,v in zip(rows, columns, cos_vals)]
-    
-    G = build_graph(rows, columns, weights, (meta[0], meta[1]))
-    return G
+        raise RuntimeError('Method not supported')
 
+    if not local:
+        G  = build_graph(rows, columns, weights, (meta[0], meta[1]))
+    else:
+        G  = build_graph_umap(rows, columns, weights, (meta[0], meta[1]), shape[1])
+    return G
 
 def la_find_partition(**kwargs):
     """
@@ -231,7 +406,7 @@ def la_find_partition(**kwargs):
 
 ###
 
-def find_communities(X, pp_method = None, nbr_num = 10, n_threads = None, verbose = False, keep_temp_files = False, temp_name = '', partition_type = la.RBConfigurationVertexPartition, resolution = 0.8, seed = None): 
+def find_communities(X, pp_method = 'None', method = 'tani', nbr_num = 10, n_threads = None, verbose = False, keep_temp_files = False, temp_name = '', partition_type = la.RBConfigurationVertexPartition, resolution = 0.8, seed = None, local = False): 
     """_summary_
 
     Parameters
@@ -267,7 +442,7 @@ def find_communities(X, pp_method = None, nbr_num = 10, n_threads = None, verbos
     RuntimeError
         System platform not supported
     """
-    X = preprocessing(X, method = pp_method)
+    X = preprocessing(X, platform = pp_method)
 
     tic = time.time()
     
@@ -318,7 +493,7 @@ def find_communities(X, pp_method = None, nbr_num = 10, n_threads = None, verbos
     
     ## readin clu file and prune with Tanimoto coefficient
     tic_prune = time.time()
-    G = build_tani_graph(knng_result, X)
+    G = build_tani_graph(knng_result, X, local = local, method = method)
     if verbose:
         print("#### Finish pruning process and build igraph graph in {0:.2f} seconds".format(time.time() - tic_prune))
 
@@ -371,23 +546,59 @@ def tanimoto_coef(x, y):
     d = np.dot(x, y)
     return d/(np.dot(x, x) + np.dot(y, y) - d)
 
-def bc(X, nbr_num = 10, seed = None, metric = "cosine"):
+def calc_jaccard(i, idx):
+    """
+    Jaccard coefficient calculation from PhenoGraph
+    :param i: index
+    :param idx: full list of neighbors in the order of index from 0
+    :return: jaccard coef
+    """
+    coefficients = np.fromiter((len(set(idx[i]).intersection(set(idx[j]))) for j in idx[i]), dtype=float)
+    coefficients /= (2 * idx.shape[1] - coefficients)
+    return coefficients
+
+
+def map_jaccard(idx):
+    """
+    map jaccard coefficient
+    :param idx: full list of neighbors in the order of index from 0
+    :return: a list of jaccard coef values in the order of index from 0 of the original list
+    """
+    n = len(idx)
+    with closing(Pool()) as pool:
+        jaccard_values = pool.starmap(calc_jaccard, zip(range(n), repeat(idx)))
+    return jaccard_values
+
+def bc(X, nbr_num = 10, seed = None, metric = "cosine", local = False, method = 'tani', pp_method = None):
+    X = preprocessing(X, platform = pp_method)
     k = nbr_num
     d, idx = bruteforce_nn.knnsearch(X, k + 1, metric)
     idx = idx[:,1:]
     shape = idx.shape
     rows = list(itertools.chain.from_iterable(itertools.repeat(i, shape[1]) for i in range(shape[0])))
     columns = list(itertools.chain.from_iterable(idx))
-    weights = map_tani(rows, columns, X)
-    G = build_graph(rows, columns, weights, (shape[0], shape[0]))
+    if method == 'tani':
+        weights = map_tani(rows, columns, X)
+    elif method =='jaccard':
+        nbr_index = np.array(idx)
+        weights = np.ravel(map_jaccard(nbr_index)).tolist()
+    elif method =='cosine':
+        weights = list(itertools.chain.from_iterable(d))
+    else:
+        raise RuntimeError('Method not supported')
+    
+    if not local:
+        G = build_graph(rows, columns, weights, (shape[0], shape[0]))
+    else:
+        G  = build_graph_umap(rows, columns, weights, (shape[0], shape[0]), shape[1])
     partition = la_find_partition(graph = G, partition_type = la.RBConfigurationVertexPartition, n_iterations = 100, resolution_parameter = 0.8, seed = seed)
     communities =  partition.membership
     return(communities)
 
 
-def clustering(X, nbr_num = 10, seed = None):
+def clustering(X, nbr_num = 10, seed = None, local = False, method =  'tani', pp_method = None):
     if sys.platform.startswith("linux"):
-        return(find_communities(X, nbr_num=nbr_num, seed = seed))
+        return(find_communities(X, nbr_num=nbr_num, seed = seed, local = local, method = method, pp_method = pp_method))
     else:
         X = preprocessing(X)
-        return(bc(X, nbr_num= nbr_num, seed = seed))
\ No newline at end of file
+        return(bc(X, nbr_num= nbr_num, seed = seed, local = local, method = method, pp_method = pp_method))
\ No newline at end of file
